
import os

from datetime import datetime, timedelta
from google.api_core.exceptions import NotFound

from airless.config import get_config
from airless.hook.google.bigquery import BigqueryHook
from airless.hook.google.storage import GcsHook
from airless.hook.file.file import FileHook
from airless.operator.base import BaseFileOperator, BaseEventOperator


class FileDetectOperator(BaseFileOperator):

    def __init__(self):
        super().__init__()
        self.gcs_hook = GcsHook()

    def execute(self, bucket, filepath):
        success_messages = self.build_success_message(bucket, filepath)

        for success_message in success_messages:
            self.pubsub_hook.publish(
                project=get_config('GCP_PROJECT'),
                topic=get_config('PUBSUB_TOPIC_FILE_TO_BQ'),
                data=success_message)

    def build_success_message(self, bucket, filepath):
        dataset, table, mode, separator, skip_leading_rows, \
            file_format, schema, run_next, quote_character, encoding, \
            column_names, time_partitioning, processing_method, \
            gcs_table_name, sheet_name, arguments = self.get_ingest_config(filepath)

        metadatas = []
        for idx in range(len(file_format)):
            metadatas.append({
                'metadata': {
                    'destination_dataset': dataset,
                    'destination_table': table,
                    'file_format': file_format[idx],
                    'mode': mode,
                    'bucket': bucket,
                    'file': filepath,
                    'separator': separator[idx],
                    'skip_leading_rows': skip_leading_rows[idx],
                    'quote_character': quote_character[idx],
                    'encoding': encoding[idx],
                    'schema': schema[idx],
                    'run_next': run_next[idx],
                    'column_names': column_names[idx],
                    'time_partitioning': time_partitioning[idx],
                    'processing_method': processing_method[idx],
                    'gcs_table_name': gcs_table_name[idx],
                    'sheet_name': sheet_name[idx],
                    'arguments': arguments[idx]
                }
            })

        return metadatas

    def get_ingest_config(self, filepath):
        dataset, table, mode = self.split_filepath(filepath)

        metadata = self.read_config_file(dataset, table)

        # Verifying if config file hava multiple configs or not
        if isinstance(metadata, list):
            metadata = metadata
        elif isinstance(metadata, dict):
            metadata = [metadata]
        else:
            raise NotImplementedError()

        # Instanciate all values
        # inputs
        file_format = []
        separator = []
        skip_leading_rows = []
        quote_character = []
        encoding = []
        sheet_name = []
        arguments = []

        # outputs
        schema = []
        column_names = []
        time_partitioning = []
        processing_method = []
        gcs_table_name = []
        run_next = []

        for config in metadata:
            # input
            file_format.append(config.get('file_format', 'csv'))
            separator.append(config.get('separator'))
            skip_leading_rows.append(config.get('skip_leading_rows'))
            quote_character.append(config.get('quote_character'))
            encoding.append(config.get('encoding', None))
            sheet_name.append(config.get('sheet_name', None))
            arguments.append(config.get('arguments', None))

            # output
            schema.append(config.get('schema', None))
            column_names.append(config.get('column_names', None))
            time_partitioning.append(config.get('time_partitioning', None))
            processing_method.append(config.get('processing_method', None))
            gcs_table_name.append(config.get('gcs_table_name', None))

            # after processing
            run_next.append(config.get('run_next', []))

        return dataset, table, mode, separator, \
            skip_leading_rows, file_format, schema, \
            run_next, quote_character, encoding, column_names, \
            time_partitioning, processing_method, gcs_table_name, \
            sheet_name, arguments

    def split_filepath(self, filepath):
        filepath_array = filepath.split('/')
        if len(filepath_array) < 3:
            raise Exception('Invalid file path. Must be added to directory {dataset}/{table}/{mode}')

        dataset = filepath_array[0]
        table = filepath_array[1]
        mode = filepath_array[2]
        return dataset, table, mode

    def read_config_file(self, dataset, table):
        try:
            config = self.gcs_hook.read_json(
                bucket=get_config('GCS_BUCKET_LANDING_ZONE_LOADER_CONFIG'),
                filepath=f'{dataset}/{table}.json')
            return config
        except NotFound:
            return {'file_format': 'json', 'time_partitioning': {'type': 'DAY', 'field': '_created_at'}}


class FileToBigqueryOperator(BaseEventOperator):

    def __init__(self):
        super().__init__()
        self.gcs_hook = GcsHook()
        self.bigquery_hook = BigqueryHook()

    def execute(self, data, topic):
        metadata = data['metadata']
        file_format = metadata['file_format']

        if file_format in ('csv', 'json'):
            self.bigquery_hook.load_file(
                from_filepath=self.gcs_hook.build_filepath(metadata['bucket'], metadata['file']),
                from_file_format=file_format,
                from_separator=metadata.get('separator'),
                from_skip_leading_rows=metadata.get('skip_leading_rows'),
                from_quote_character=metadata.get('quote_character'),
                from_encoding=metadata.get('encoding'),
                to_project=get_config('GCP_PROJECT'),
                to_dataset=metadata['destination_dataset'],
                to_table=metadata['destination_table'],
                to_mode=metadata['mode'],
                to_schema=metadata.get('schema'),
                to_time_partitioning=metadata.get('time_partitioning'))

        else:
            raise Exception(f'File format {file_format} load not implemented')


class BatchWriteDetectOperator(BaseEventOperator):

    def __init__(self):
        super().__init__()
        self.file_hook = FileHook()
        self.gcs_hook = GcsHook()

    def execute(self, data, topic):
        bucket = data.get('bucket', get_config('GCS_BUCKET_LANDING_ZONE'))
        threshold = data['threshold']

        tables = {}
        partially_processed_tables = []

        for b in self.gcs_hook.list(bucket):
            if b.time_deleted is None:
                filepaths = b.name.split('/')
                key = '/'.join(filepaths[:-1])  # dataset/table
                filename = filepaths[-1]

                if tables.get(key) is None:
                    tables[key] = {
                        'size': b.size,
                        'files': [filename],
                        'min_time_created': b.time_created
                    }
                else:
                    tables[key]['size'] += b.size
                    tables[key]['files'] += [filename]
                    if b.time_created < tables[key]['min_time_created']:
                        tables[key]['min_time_created'] = b.time_created

                if (tables[key]['size'] > threshold['size']) or (len(tables[key]['files']) > threshold['file_quantity']):
                    self.send_to_process(bucket=bucket, directory=key, files=tables[key]['files'])
                    tables[key] = None
                    partially_processed_tables.append(key)

        # verify which dataset/table is ready to be processed
        time_threshold = (datetime.now() - timedelta(minutes=threshold['minutes'])).strftime('%Y-%m-%d %H:%M')
        for directory, v in tables.items():
            if v is not None:
                if (v['size'] > threshold['size']) or \
                    (v['min_time_created'].strftime('%Y-%m-%d %H:%M') < time_threshold) or \
                        (len(v['files']) > threshold['file_quantity']) or \
                        (directory in partially_processed_tables):
                    self.send_to_process(bucket=bucket, directory=directory, files=v['files'])

    def send_to_process(self, bucket, directory, files):
        self.pubsub_hook.publish(
            project=get_config('GCP_PROJECT'),
            topic=get_config('PUBSUB_TOPIC_BATCH_WRITE_PROCESS'),
            data={'bucket': bucket, 'directory': directory, 'files': files})


class BatchWriteProcessOperator(BaseEventOperator):

    def __init__(self):
        super().__init__()
        self.file_hook = FileHook()
        self.gcs_hook = GcsHook()

    def execute(self, data, topic):
        from_bucket = data['bucket']
        directory = data['directory']
        files = data['files']

        file_contents = self.read_files(from_bucket, directory, files)

        local_filepath = self.merge_files(file_contents)

        self.gcs_hook.upload(local_filepath, get_config('GCS_BUCKET_LANDING_ZONE_LOADER'), f'{directory}/append')
        os.remove(local_filepath)

        self.move_files(from_bucket, get_config('GCS_BUCKET_LANDING_ZONE_PROCESSED'), directory, files)

    def read_files(self, bucket, directory, files):
        file_contents = []
        for f in files:
            obj = self.gcs_hook.read_json(
                bucket=bucket,
                filepath=f'{directory}/{f}')
            if isinstance(obj, list):
                file_contents += obj
            elif isinstance(obj, dict):
                file_contents.append(obj)
            else:
                raise Exception(f'Cannot process file {directory}/{f}')
        return file_contents

    def merge_files(self, file_contents):
        local_filepath = self.file_hook.get_tmp_filepath('merged.ndjson', add_timestamp=True)
        self.file_hook.write(local_filepath=local_filepath, data=file_contents, use_ndjson=True)
        return local_filepath

    def move_files(self, from_bucket, to_bucket, directory, files):
        for f in files:
            self.gcs_hook.move(
                from_bucket=from_bucket,
                from_prefix=f'{directory}/{f}',
                to_bucket=to_bucket,
                to_directory=directory)


class FileDeleteOperator(BaseEventOperator):

    def __init__(self):
        super().__init__()
        self.gcs_hook = GcsHook()

    def execute(self, data, topic):
        bucket = data['bucket']
        prefix = data['prefix']
        self.gcs_hook.delete(bucket, prefix)


class FileMoveOperator(BaseEventOperator):

    def __init__(self):
        super().__init__()
        self.gcs_hook = GcsHook()

    def execute(self, data, topic):
        origin_bucket = data['origin']['bucket']
        origin_prefix = data['origin']['prefix']
        dest_bucket = data['destination']['bucket']
        dest_directory = data['destination']['directory']
        self.gcs_hook.move(origin_bucket, origin_prefix, dest_bucket, dest_directory, True)




import os

from airless.hook.file.ftp import FtpHook
from airless.operator.file.file import FileUrlToGcsOperator


class FtpToGcsOperator(FileUrlToGcsOperator):

    def __init__(self):
        super().__init__()
        self.ftp_hook = FtpHook()

    def execute(self, data, topic):
        origin = data['origin']
        destination = data['destination']

        self.ftp_hook = FtpHook()
        self.ftp_hook.login(origin['host'], origin.get('user'), origin.get('pass'))

        local_filepath = self.ftp_hook.download(origin['directory'], origin['filename'])

        self.move_to_destinations(local_filepath, destination)

        os.remove(local_filepath)




import json
import time

from datetime import datetime

from airless.config import get_config
from airless.dto.pubsub_to_bq import PubsubToBqDto
from airless.operator.base import BaseEventOperator


class ErrorReprocessOperator(BaseEventOperator):

    def __init__(self):
        super().__init__()

    def execute(self, data, topic):

        input_type = data['input_type']
        origin = data.get('origin', 'undefined')
        message_id = data.get('event_id')
        original_data = data['data']
        metadata = original_data.get('metadata', {})

        retry_interval = metadata.get('retry_interval', 5)
        retries = metadata.get('retries', 0)
        max_retries = metadata.get('max_retries', 2)

        if (input_type == 'event') and (retries < max_retries):
            time.sleep(min(retry_interval ** retries, 480))  # Cloud function max execution time is 540s (9 min), so set it to wait max 480s (8 min)
            original_data.setdefault('metadata', {})['retries'] = retries + 1
            self.pubsub_hook.publish(
                project=get_config('GCP_PROJECT'),
                topic=origin,
                data=original_data)

        else:
            dto = PubsubToBqDto(
                event_id=message_id,
                resource=origin,
                to_project=get_config('GCP_PROJECT'),
                to_dataset=get_config('BIGQUERY_DATASET_ERROR'),
                to_table=get_config('BIGQUERY_TABLE_ERROR'),
                to_schema=None,
                to_partition_column='_created_at',
                to_extract_to_cols=False,
                to_keys_format=None,
                data=data)
            self.pubsub_hook.publish(
                project=get_config('GCP_PROJECT'),
                topic=get_config('PUBSUB_TOPIC_PUBSUB_TO_BQ'),
                data=dto.as_dict())

            email_send_topic = get_config('PUBSUB_TOPIC_EMAIL_SEND', False)
            if email_send_topic and (origin != email_send_topic):
                self.notify_email(origin, message_id, data)

            slack_send_topic = get_config('PUBSUB_TOPIC_SLACK_SEND', False)
            if slack_send_topic and (origin != slack_send_topic):
                self.notify_slack(origin, message_id, data)

    def notify_email(self, origin, message_id, data):
        email_message = {
            'sender': get_config('EMAIL_SENDER_ERROR'),
            'recipients': eval(get_config('EMAIL_RECIPIENTS_ERROR')),
            'subject': f'{origin} | {message_id}',
            'content': f'Input Type: {data["input_type"]} Origin: {origin}\nMessage ID: {message_id}\n\n {json.dumps(data["data"])}\n\n{data["error"]}'
        }
        self.pubsub_hook.publish(
            project=get_config('GCP_PROJECT'),
            topic=get_config('PUBSUB_TOPIC_EMAIL_SEND'),
            data=email_message)

    def notify_slack(self, origin, message_id, data):
        slack_message = {
            'channels': eval(get_config('SLACK_CHANNELS_ERROR')),
            'message': f'{origin} | {message_id}\n\n{json.dumps(data["data"])}\n\n{data["error"]}'
        }
        self.pubsub_hook.publish(
            project=get_config('GCP_PROJECT'),
            topic=get_config('PUBSUB_TOPIC_SLACK_SEND'),
            data=slack_message)

    def prepare_row(self, row, message_id, origin):
        return {
            '_event_id': message_id or 1234,
            '_resource': origin or 'local',
            '_json': json.dumps(row),
            '_created_at': str(datetime.now())
        }

    def prepare_rows(self, data, message_id, origin):
        prepared_rows = data if isinstance(data, list) else [data]
        return [self.prepare_row(row, message_id, origin) for row in prepared_rows]




import json

from google.cloud import pubsub_v1

from airless.hook.base import BaseHook
from airless.config import get_config


class PubsubHook(BaseHook):

    def __init__(self):
        super().__init__()
        self.publisher = pubsub_v1.PublisherClient()

    def publish(self, project, topic, data):

        if get_config('ENV') == 'prod':
            topic_path = self.publisher.topic_path(project, topic)

            message_bytes = json \
                .dumps(data) \
                .encode('utf-8')

            publish_future = self.publisher.publish(topic_path, data=message_bytes)
            publish_future.result(timeout=10)
            self.logger.info(f'published to {project}.{topic}')
            return 'Message published.'
        else:
            self.logger.debug(f'[DEV] Message published to Project {project}, Topic {topic}: {data}')




import os
import re

from airless.hook.file.file import FileHook
from airless.hook.google.storage import GcsHook
from airless.operator.base import BaseEventOperator


class FileUrlToGcsOperator(BaseEventOperator):

    def __init__(self):
        super().__init__()
        self.file_hook = FileHook()
        self.gcs_hook = GcsHook()

    def execute(self, data, topic):
        origin = data['origin']
        destination = data['destination']

        local_filepath = self.file_hook.download(
            url=origin['url'],
            headers=origin.get('headers'),
            timeout=origin.get('timeout', 500),
            proxies=origin.get('proxies')
        )

        self.move_to_destinations(local_filepath, destination)

        os.remove(local_filepath)

    def move_to_destinations(self, local_filepath, destination):

        original_filepath = local_filepath
        destinations = destination if isinstance(destination, list) else [destination]

        for dest in destinations:

            if dest.get('filename'):
                local_filepath = self.file_hook.rename(
                    from_filename=local_filepath,
                    to_filename=dest.get('filename'))

            bucket = dest['bucket']
            directory = dest.get('directory', f"{dest.get('dataset')}/{dest.get('table')}/{dest.get('mode')}")
            remove_null_byte = dest.get('remove_null_byte')
            regex = dest.get('regex', '.*')

            if re.search(regex, local_filepath, re.IGNORECASE):

                if remove_null_byte:
                    self.remove_null_byte(local_filepath)
                self.gcs_hook.upload(local_filepath, bucket, directory)

                if local_filepath != original_filepath:  # revert to original filename
                    local_filepath = self.file_hook.rename(
                        from_filename=local_filepath,
                        to_filename=original_filepath)

    def remove_null_byte(self, local_filepath):
        splits = local_filepath.split('/')
        tmp_file = '/'.join(splits[:-1] + ['tmp-' + splits[-1]])

        escaped_filepath = re.escape(local_filepath)

        response = os.system(f"tr -d '\\000' < {escaped_filepath} > {tmp_file} && mv {tmp_file} {escaped_filepath}")
        if response != 0:
            raise Exception('not able to remove null bytes')




import json
import re

from datetime import datetime
from unidecode import unidecode

from airless.config import get_config
from airless.dto.pubsub_to_bq import PubsubToBqDto
from airless.hook.google.bigquery import BigqueryHook
from airless.hook.google.storage import GcsHook
from airless.operator.base import BaseEventOperator


class GcsQueryToBigqueryOperator(BaseEventOperator):

    def __init__(self):
        super().__init__()

        self.gcs_hook = GcsHook()
        self.bigquery_hook = BigqueryHook()

    def execute(self, data, topic):

        query = data['query']
        if isinstance(query, dict):
            query_bucket = query.get('bucket', get_config('GCS_BUCKET_SQL'))
            query_filepath = query['filepath']
            query_params = query.get('params', {})
        else:
            query_bucket = get_config('GCS_BUCKET_SQL')
            query_filepath = query
            query_params = data.get('params', {})

        to = data.get('to', {})

        if to:
            to_project = to.get('project', get_config('GCP_PROJECT'))
            to_dataset = to.get('dataset')
            to_table = to.get('table')
            to_write_disposition = to.get('write_disposition')
            to_time_partitioning = to.get('time_partitioning')
        else:
            to_project = get_config('GCP_PROJECT')
            to_dataset = data.get('destination_dataset')
            to_table = data.get('destination_table')
            to_write_disposition = data.get('write_disposition')
            to_time_partitioning = data.get('time_partitioning')

        sql = self.gcs_hook.read(query_bucket, query_filepath, 'utf-8')
        for k, v in query_params.items():
            sql = sql.replace(f':{k}', str(v))

        self.bigquery_hook.execute_query_job(
            sql, to_project, to_dataset,
            to_table, to_write_disposition, to_time_partitioning,
            timeout=float(get_config('BIGQUERY_JOB_TIMEOUT', False, 480)))


class PubsubToBqOperator(BaseEventOperator):

    def __init__(self):
        super().__init__()
        self.bigquery_hook = BigqueryHook()

    def execute(self, data, topic):
        dto = PubsubToBqDto.from_dict(data)

        prepared_rows = self.prepare_rows(dto)

        self.bigquery_hook.write(
            project=dto.to_project,
            dataset=dto.to_dataset,
            table=dto.to_table,
            schema=dto.to_schema,
            partition_column=dto.to_partition_column,
            rows=prepared_rows)

    def prepare_row(self, row, event_id, resource, extract_to_cols, keys_format):
        prepared_row = {
            '_event_id': event_id,
            '_resource': resource,
            '_json': json.dumps(row),
            '_created_at': str(datetime.now())
        }

        if extract_to_cols:
            for key in row.keys():
                if (key not in ['_event_id', '_resource', '_json', '_created_at']) and (row[key] is not None):
                    new_key = key
                    if keys_format == 'lowercase':
                        new_key = key.lower()
                        new_key = self.format_key(new_key)
                    elif keys_format == 'snakecase':
                        new_key = self.camel_to_snake(key)
                        new_key = self.format_key(new_key)

                    if isinstance(row[key], list) or isinstance(row[key], dict):
                        prepared_row[new_key] = json.dumps(row[key])
                    else:
                        prepared_row[new_key] = str(row[key])

        return prepared_row

    def prepare_rows(self, dto):
        prepared_rows = dto.data if isinstance(dto.data, list) else [dto.data]
        return [self.prepare_row(row, dto.event_id, dto.resource, dto.to_extract_to_cols, dto.to_keys_format) for row in prepared_rows]

    def camel_to_snake(self, s):
        return ''.join(['_' + c.lower() if c.isupper() else c for c in s]).lstrip('_')

    def format_key(self, key):
        return re.sub(r'[^a-z0-9_]', '', unidecode(key.lower().replace(' ', '_')))




from airless.config import get_config
from airless.operator.base import BaseEventOperator
from airless.hook.google.secret_manager import SecretManagerHook
from airless.hook.notification.slack import SlackHook


class SlackSendOperator(BaseEventOperator):

    def __init__(self):
        super().__init__()
        self.slack_hook = SlackHook()
        self.secret_manager_hook = SecretManagerHook()

    def execute(self, data, topic):
        channels = data['channels']
        secret_id = data.get('secret_id', 'slack_alert')
        message = data.get('message')
        blocks = data.get('blocks')
        thread_ts = data.get('thread_ts')
        reply_broadcast = data.get('reply_broadcast', False)

        token = self.secret_manager_hook.get_secret(get_config('GCP_PROJECT'), secret_id, True)['bot_token']
        self.slack_hook.set_token(token)

        for channel in channels:
            self.slack_hook.send(channel, message, blocks, thread_ts, reply_broadcast)


class SlackReactOperator(BaseEventOperator):

    def __init__(self):
        super().__init__()
        self.slack_hook = SlackHook()
        self.secret_manager_hook = SecretManagerHook()

    def execute(self, data, topic):
        channel = data['channel']
        secret_id = data.get('secret_id', 'slack_alert')
        reaction = data.get('reaction')
        ts = data.get('ts')

        token = self.secret_manager_hook.get_secret(get_config('GCP_PROJECT'), secret_id, True)['bot_token']
        self.slack_hook.set_token(token)

        self.slack_hook.react(channel, reaction, ts)




import json
import ndjson
import os
import requests

from datetime import datetime
from random import randint

from airless.hook.base import BaseHook


class FileHook(BaseHook):

    def __init__(self):
        super().__init__()

    def write(self, local_filepath, data, use_ndjson=False):
        with open(local_filepath, 'w') as f:
            if isinstance(data, dict) or isinstance(data, list):
                if use_ndjson:
                    ndjson.dump(data, f)
                else:
                    json.dump(data, f)
            else:
                f.write(str(data))

    def extract_filename(self, filepath_or_url):
        return filepath_or_url.split('/')[-1].split('?')[0].split('#')[0]

    def get_tmp_filepath(self, filepath_or_url, add_timestamp=True):
        filename = self.extract_filename(filepath_or_url)
        if add_timestamp:
            timestamp = datetime.now().strftime('%Y%m%d%H%M%S')
            filename = f'{timestamp}_{randint(1, 100000000)}_{filename}'
        return f'/tmp/{filename}'

    def download(self, url, headers, timeout=500, proxies=None):
        local_filename = self.get_tmp_filepath(url)
        with requests.get(url, stream=True, verify=False, headers=headers, timeout=timeout, proxies=proxies) as r:
            r.raise_for_status()
            with open(local_filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        return local_filename

    def rename(self, from_filename, to_filename):
        to_filename_formatted = ('' if to_filename.startswith('/tmp/') else '/tmp/') + to_filename
        os.rename(from_filename, to_filename_formatted)
        return to_filename_formatted

    def rename_files(self, dir, prefix):
        for root, subdirs, files in os.walk(dir):
            for filename in files:
                os.rename(os.path.join(root, filename), os.path.join(root, f'{prefix}_{filename}'))

    def list_files(self, folder):
        file_list = []
        for root, subdirs, files in os.walk(folder):
            for filename in files:
                filepath = os.path.join(root, filename)
                file_list.append(filepath)

        return file_list




from airless.operator.base import BaseEventOperator
from time import sleep


class DelayOperator(BaseEventOperator):

    """
    Operator that adds a delay to the pipeline. The maximum delay is 500 due
    to Cloud Functions time constraint of 540 of execution time

    It can receive 1 parameter:
    seconds: number of seconds to wait

    """

    def __init__(self):
        super().__init__()

    def execute(self, data, topic):
        seconds = data['seconds']
        seconds = min(seconds, 500)
        sleep(seconds)




import re

from dateutil import parser
from ftplib import FTP

from airless.hook.file.file import FileHook


class FtpHook(FileHook):

    def __init__(self):
        super().__init__()
        self.ftp = None

    def login(self, host, user, password):
        self.ftp = FTP(host, user, password)
        self.ftp.login()

    def cwd(self, dir):
        if dir:
            self.ftp.cwd(dir)

    def list(self, regex=None, updated_after=None, updated_before=None):
        lines = []
        self.ftp.dir("", lines.append)

        files = []
        directories = []

        for line in lines:
            tokens = line.split()
            obj = {
                'name': tokens[3],
                'updated_at': parser.parse(' '.join(tokens[:1]))
            }

            if regex and not re.search(regex, obj['name'], re.IGNORECASE):
                continue

            if updated_after and not (obj['updated_at'] >= updated_after):
                continue

            if updated_before and not (obj['updated_at'] <= updated_before):
                continue

            obj = {
                'name': tokens[3],
                'updated_at': parser.parse(' '.join(tokens[:1]))
            }
            if tokens[2] == '<DIR>':
                directories.append(obj)
            else:
                files.append(obj)

        return files, directories

    def download(self, dir, filename):
        self.cwd(dir)
        local_filepath = self.get_tmp_filepath(filename)
        with open(local_filepath, 'wb') as file:
            self.ftp.retrbinary(f'RETR {filename}', file.write)
        return local_filepath




import json
import logging
import time
import traceback

from base64 import b64decode

from airless.config import get_config
from airless.hook.google.pubsub import PubsubHook


class BaseOperator():

    def __init__(self):
        self.logger = logging.getLogger(f'{self.__class__.__module__}.{self.__class__.__name__}')
        logging.basicConfig(level=logging.getLevelName(get_config('LOG_LEVEL')))
        logging.debug(f'Created class instance {self.__class__.__name__}')
        self.pubsub_hook = PubsubHook()
        self.trigger_type = None
        self.message_id = None

    def extract_message_id(self, cloud_event):
        return cloud_event['id']

    def report_error(self, message, data=None):
        if get_config('ENV') == 'prod':
            logging.error(f'Error {message}')
        else:
            logging.debug(f'[DEV] Error {message}')

        error_obj = self.build_error_message(message, data)
        self.pubsub_hook.publish(
            project=get_config('GCP_PROJECT'),
            topic=get_config('PUBSUB_TOPIC_ERROR'),
            data=error_obj)

    def build_error_message(self, message, data):
        raise NotImplementedError()

    def chain_messages(self, messages):
        msg_chain = None
        messages.reverse()

        for m in messages:
            new_msg = m['data'].copy()
            if msg_chain:
                new_msg['metadata'] = {**new_msg.get('metadata', {}), **msg_chain}

            msg_chain = {
                'run_next': [{
                    'topic': m['topic'],
                    'data': new_msg
                }]
            }

        chained_messages = msg_chain['run_next'][0]['data']
        first_topic = msg_chain['run_next'][0]['topic']

        return chained_messages, first_topic


class BaseFileOperator(BaseOperator):

    def __init__(self):
        super().__init__()

        self.trigger_type = 'file'
        self.trigger_origin = None
        self.cloud_event = None

    def execute(self, bucket, filepath):
        raise NotImplementedError()

    def run(self, cloud_event):
        self.logger.debug(cloud_event)
        try:
            self.message_id = self.extract_message_id(cloud_event)
            self.cloud_event = cloud_event
            trigger_file_bucket = cloud_event['bucket']
            trigger_file_path = cloud_event.data['name']
            self.trigger_origin = f'{trigger_file_bucket}/{trigger_file_path}'
            self.execute(trigger_file_bucket, trigger_file_path)

        except Exception as e:
            self.report_error(f'{str(e)}\n{traceback.format_exc()}')

    def build_error_message(self, message, data):
        return {
            'input_type': self.trigger_type,
            'origin': self.trigger_origin,
            'error': message,
            'event_id': self.message_id,
            'data': {
                'attributes': self.cloud_event._attributes,
                'data': data or self.cloud_event.data
            }
        }


class BaseEventOperator(BaseOperator):

    def __init__(self):
        super().__init__()

        self.trigger_type = 'event'
        self.trigger_event_topic = None
        self.trigger_event_data = None

    def execute(self, data, topic):
        raise NotImplementedError()

    def run(self, cloud_event):
        self.logger.debug(cloud_event)
        try:
            self.message_id = self.extract_message_id(cloud_event)
            decoded_data = b64decode(cloud_event.data['message']['data']).decode('utf-8')
            self.trigger_event_data = json.loads(decoded_data)
            self.trigger_event_topic = cloud_event['source'].split('/')[-1]

            self.execute(self.trigger_event_data, self.trigger_event_topic)

            tasks = self.trigger_event_data.get('metadata', {}).get('run_next', [])
            self.run_next(tasks)

        except Exception as e:
            self.report_error(f'{str(e)}\n{traceback.format_exc()}')

    def run_next(self, tasks):
        if tasks:
            time.sleep(10)
        for t in tasks:
            self.pubsub_hook.publish(
                project=t.get('project', get_config('GCP_PROJECT')),
                topic=t['topic'],
                data=t['data'])

    def build_error_message(self, message, data):
        return {
            'input_type': self.trigger_type,
            'origin': self.trigger_event_topic,
            'error': message,
            'event_id': self.message_id,
            'data': data or self.trigger_event_data
        }


class BaseHttpOperator(BaseOperator):

    def __init__(self):
        super().__init__()

        self.trigger_type = 'http'
        self.trigger_base_url = None
        self.trigger_request = None

    def execute(self, request):
        raise NotImplementedError()

    def run(self, request):
        self.logger.debug(request)
        try:
            self.trigger_request = {
                'url': request.base_url,
                'method': request.method,
                'form': request.form.to_dict(),
                'args': request.args.to_dict(),
                'data': request.data.decode('utf-8')
            }
            self.trigger_base_url = request.base_url

            self.execute(request)

        except Exception as e:
            self.report_error(f'{str(e)}\n{traceback.format_exc()}')

    def build_error_message(self, message, request):
        return {
            'input_type': self.trigger_type,
            'origin': self.trigger_base_url,
            'error': message,
            'event_id': int(time.time() * 1000),
            'data': request or self.trigger_request
        }




import json
import os

from datetime import datetime
from google.cloud import storage
from itertools import zip_longest

from airless.config import get_config
from airless.hook.base import BaseHook
from airless.hook.file.file import FileHook


def natatime(n, iterable, fillvalue=None):
    stepped_slices = [iter(iterable)] * n
    return zip_longest(*stepped_slices, fillvalue=fillvalue)


class GcsHook(BaseHook):

    def __init__(self):
        super().__init__()
        self.storage_client = storage.Client()
        self.file_hook = FileHook()

    def build_filepath(self, bucket, filepath):
        return f'gs://{bucket}/{filepath}'

    def read(self, bucket, filepath, encoding=None):
        bucket = self.storage_client.get_bucket(bucket)

        blob = bucket.blob(filepath)
        content = blob.download_as_string()
        if encoding:
            return content.decode(encoding)
        else:
            return content.decode()

    def download(self, bucket, filepath, target_filepath=None):
        bucket = self.storage_client.get_bucket(bucket)

        filename = filepath.split('/')[-1]
        blob = bucket.blob(filepath)
        blob.download_to_filename(target_filepath or filename)

    def read_json(self, bucket, filepath, encoding=None):
        return json.loads(self.read(bucket, filepath, encoding))

    def upload_from_memory(self, data, bucket, directory, filename, add_timestamp):
        local_filename = self.file_hook.get_tmp_filepath(filename, add_timestamp)
        try:
            self.file_hook.write(local_filename, data)
            self.upload(local_filename, bucket, directory)

        finally:
            if os.path.exists(local_filename):
                os.remove(local_filename)

    def upload(self, local_filepath, bucket, directory):
        filename = self.file_hook.extract_filename(local_filepath)
        bucket = self.storage_client.bucket(bucket)
        blob = bucket.blob(f"{directory}/{filename}")
        blob.upload_from_filename(local_filepath)

    def check_existance(self, bucket, filepath):
        blobs = self.storage_client.list_blobs(bucket, prefix=filepath, max_results=1, page_size=1)
        return len(list(blobs)) > 0

    def move(self, from_bucket, from_prefix, to_bucket, to_directory, rewrite=False):
        bucket = self.storage_client.get_bucket(from_bucket)
        blobs = bucket.list_blobs(prefix=from_prefix)

        dest_bucket = self.storage_client.bucket(to_bucket)

        for blob in blobs:
            if not blob.name.endswith('/'):
                filename = blob.name.split('/')[-1]

                if rewrite:
                    rewrite_token = False
                    dest_blob = dest_bucket.blob(f'{to_directory}/{filename}')
                    while True:
                        rewrite_token, bytes_rewritten, bytes_to_rewrite = dest_blob.rewrite(
                            blob, token=rewrite_token)
                        self.logger.debug(f'{to_directory}/{filename} - Progress so far: {bytes_rewritten}/{bytes_to_rewrite} bytes')

                        if not rewrite_token:
                            break

                else:
                    bucket.copy_blob(
                        blob, dest_bucket, f'{to_directory}/{filename}'
                    )
                bucket.delete_blob(blob.name)

    def delete(self, bucket_name, prefix):
        bucket = self.storage_client.get_bucket(bucket_name)
        blobs = bucket.list_blobs(prefix=prefix)

        for to_delete_batch in natatime(100, blobs):
            tmp_list = [tdb for tdb in to_delete_batch if (tdb is not None) and (tdb.name != prefix)]
            if len(tmp_list) > 0:
                with self.storage_client.batch():
                    for blob in tmp_list:
                        blob.delete()

    def list(self, bucket_name):
        return self.storage_client.list_blobs(bucket_name, fields='items(name,size,timeCreated,timeDeleted),nextPageToken')


class GcsDatalakeHook(GcsHook):

    def __init__(self):
        super().__init__()

    def build_metadata(self, message_id, origin):
        return {
            'event_id': message_id or 1234,
            'resource': origin or 'local'
        }

    def prepare_row(self, row, metadata):
        return {
            '_event_id': metadata['event_id'],
            '_resource': metadata['resource'],
            '_json': json.dumps({'data': row, 'metadata': metadata}),
            '_created_at': str(datetime.now())
        }

    def prepare_rows(self, data, metadata):
        prepared_rows = data if isinstance(data, list) else [data]
        return [self.prepare_row(row, metadata) for row in prepared_rows]

    def send_to_landing_zone(self, data, dataset, table, message_id, origin):

        if isinstance(data, list) and (len(data) == 0):
            raise Exception(f'Trying to send empty list to landing zone: {dataset}.{table}')

        if isinstance(data, dict) and (data == {}):
            raise Exception(f'Trying to send empty dict to landing zone: {dataset}.{table}')

        if get_config('ENV') == 'prod':
            metadata = self.build_metadata(message_id, origin)
            prepared_rows = self.prepare_rows(data, metadata)
            self.upload_from_memory(
                data=prepared_rows,
                bucket=get_config('GCS_BUCKET_LANDING_ZONE'),
                directory=f'{dataset}/{table}',
                filename='tmp.json',
                add_timestamp=True)
        else:
            self.logger.debug(f'[DEV] Uploading to {dataset}.{table}, Data: {data}')




import requests

from airless.hook.base import BaseHook


class SlackHook(BaseHook):

    def __init__(self):
        super().__init__()
        self.api_url = 'slack.com'

    def set_token(self, token):
        self.token = token

    def get_headers(self):
        return {
            'Authorization': f'Bearer {self.token}'
        }

    def send(self, channel, message=None, blocks=None, thread_ts=None, reply_broadcast=False):

        data = {
            'channel': channel,
            'text': message
        }

        if message:
            message = message[:3000]  # slack does not accept long messages
            data['text'] = message

        if blocks:
            data['blocks'] = blocks

        if thread_ts:
            data['thread_ts'] = thread_ts
            data['reply_broadcast'] = reply_broadcast

        response = requests.post(
            f'https://{self.api_url}/api/chat.postMessage',
            headers=self.get_headers(),
            json=data,
            timeout=10
        )
        response.raise_for_status()

    def react(self, channel, reaction, ts):
        data = {
            'channel': channel,
            'name': reaction,
            'timestamp': ts
        }
        response = requests.post(
            f'https://{self.api_url}/api/reactions.add',
            headers=self.get_headers(),
            json=data,
            timeout=10
        )
        response.raise_for_status()



from airless.operator.base import BaseEventOperator
from airless.hook.notification.email import EmailHook
from airless.hook.google.storage import GcsHook


class EmailSendOperator(BaseEventOperator):

    def __init__(self):
        super().__init__()
        self.email_hook = EmailHook()
        self.gcs_hook = GcsHook()

    def execute(self, data, topic):
        subject = data['subject']
        content = data['content']
        recipients = data['recipients']
        sender = data.get('sender', 'Airless notification')
        attachments = data.get('attachments', [])
        mime_type = data.get('mime_type', 'plain')

        attachment_contents = []
        for att in attachments:
            attachment_contents.append({
                'type': att.get('type', 'text'),
                'content': self.gcs_hook.read(att['bucket'], att['filepath'], att['encoding'])
            })

        self.email_hook.send(subject, content, recipients, sender, attachment_contents, mime_type)
